'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol ? "symbol" : typeof obj; };

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

var INIT_REDUX = '@@redux/INIT';
var INIT_DEVTOOLS = '@@INIT';
var INIT_REDUX_OPERATIONS = exports.INIT_REDUX_OPERATIONS = '@@reduxOperations/INIT';
var REDUX_OPERATION_SIGNATURE = exports.REDUX_OPERATION_SIGNATURE = '@@reduxOperations';

var walkState = exports.walkState = function walkState() {
  var locationStack = arguments.length <= 0 || arguments[0] === undefined ? [] : arguments[0];
  var state = arguments[1];
  var initializer = arguments[2];

  return locationStack.reduce(function (reduction, key, currentIdx) {
    if (!reduction[key]) {
      reduction[key] = currentIdx === locationStack.length - 1 ? undefined : {};
    }
    return reduction[key] || initializer && initializer(undefined, { type: INIT_REDUX });
  }, state);
};

var appendChangeToState = function appendChangeToState(locationStack, state, newSubState) {
  if (locationStack.length === 1) {
    return Object.assign({}, state, _defineProperty({}, locationStack[0], newSubState));
  } else {
    var subObject = appendChangeToState(locationStack.slice(1), state[locationStack[0]], newSubState);
    return Object.assign({}, state, _defineProperty({}, locationStack[0], subObject));
  }
};

var makeStoreAPI = function makeStoreAPI(initResult) {
  var api = {};
  makeStoreOperations(api, initResult);
  Object.keys(api).forEach(function (action) {
    api[action].operationArray.sort(function (a, b) {
      var priorityA = a.priority || 0;
      var priorityB = b.priority || 0;
      return priorityA > priorityB;
    });
    if (process.env.NODE_ENV !== 'production') {
      api[action].arguments = api[action].operationArray.reduce(function (reduction, operation) {
        return Object.assign(reduction, operation.arguments);
      }, {});
    }
  });
  return api;
};

var makeStoreOperations = function makeStoreOperations(storeOperations, state) {
  var stack = arguments.length <= 2 || arguments[2] === undefined ? [] : arguments[2];
  var key = arguments[3];

  if (state && (typeof state === 'undefined' ? 'undefined' : _typeof(state)) === 'object' && state.signature === REDUX_OPERATION_SIGNATURE) {
    Object.keys(state).filter(function (key) {
      return key !== 'signature';
    }).forEach(function (operation) {
      storeOperations[operation] = storeOperations[operation] || {};
      storeOperations[operation].operationArray = storeOperations[operation].operationArray || [];
      storeOperations[operation].operationArray.push(_extends({}, state[operation], {
        defaultLocation: [].concat(_toConsumableArray(stack)),
        name: key
      }));
    });
  } else {
    Object.keys(state).forEach(function (key) {
      stack.push(key);
      makeStoreOperations(storeOperations, state[key], stack, key);
    });
  }
  stack.pop();
};

var liftReducerWith = function liftReducerWith(reducer, initialCommittedState) {
  var initialLiftedState = {
    userState: initialCommittedState
  };

  return function () {
    var liftedState = arguments.length <= 0 || arguments[0] === undefined ? initialLiftedState : arguments[0];
    var liftedAction = arguments[1];
    var api = liftedState.api;
    var userState = liftedState.userState;

    var activeState = reducer(userState, liftedAction);
    if (liftedAction.type === INIT_REDUX || liftedAction.type === INIT_DEVTOOLS) {
      var initResult = reducer(undefined, { type: INIT_REDUX_OPERATIONS });
      api = makeStoreAPI(initResult);
    } else {
      var actionObject = api[liftedAction.type] || {};
      var operationArray = actionObject.operationArray;
      if (operationArray) {
        operationArray.forEach(function (operation) {
          var locationStack = operation.defaultLocation;
          // 3 possiblies: If a locationStack isn't given, use the default (for simple non-multi scenarios)
          // If Loc but no Name, or name == operation name, use given location (for dynamic or multi scenarios)
          // Otherwise, use default
          if (liftedAction.meta.location && (!liftedAction.meta.name || operation.name === liftedAction.meta.name)) {
            locationStack = liftedAction.meta.location;
          }
          var subState = walkState(locationStack, activeState);
          var newSubState = operation.resolve(subState, liftedAction);
          if (subState !== newSubState) {
            activeState = appendChangeToState(locationStack, activeState, newSubState);
          }
          liftedAction.meta.operationResults = liftedAction.meta.operationResults || {};
          liftedAction.meta.operationResults[operation.name] = { oldState: subState, state: newSubState };
        });
      }
    }
    return {
      api: api,
      userState: activeState
    };
  };
};

var unliftState = function unliftState(liftedState) {
  return liftedState.userState;
};

var unliftStore = function unliftStore(reduxOperationStore, liftReducer) {
  return _extends({}, reduxOperationStore, {
    dispatch: function dispatch(action) {
      action.meta = action.meta || {};
      action.meta.dispatch = reduxOperationStore.dispatch;
      reduxOperationStore.dispatch(action);
      return action;
    },
    getState: function getState() {
      return unliftState(reduxOperationStore.getState());
    },
    replaceReducer: function replaceReducer(nextReducer) {
      reduxOperationStore.replaceReducer(liftReducer(nextReducer));
    }
  });
};

var reduxOperations = exports.reduxOperations = function reduxOperations() {
  return function (createStore) {
    return function (reducer, initialState, enhancer) {
      function liftReducer(r) {
        if (typeof r !== 'function') {
          throw new Error('Expected the reducer to be a function.');
        }
        return liftReducerWith(r, initialState);
      }
      var reduxOperationStore = createStore(liftReducer(reducer), enhancer);
      if (reduxOperationStore.reduxOperationStore) {
        throw new Error('reduxOperation should not be applied more than once. Check your store configuration.');
      }
      return unliftStore(reduxOperationStore, liftReducer);
    };
  };
};